/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */
/*                              Contornos                                  */
/* ----------------------------------------------------------------------- */

#include <stdlib.h>
#include <math.h>

#include "IMG_CONT.H"

/* ----------------------------------------------------------------------- */
/*  Orientacin CCW                                                        */

static void PixelAdyacente (int x, int y, int orientacion, int *sx, int *sy)
{
  switch (orientacion) {

    case 0: *sx = x+1;
            *sy = y;
            break;

    case 1: *sx = x+1;
            *sy = y-1;
            break;

    case 2: *sx = x;
            *sy = y-1;
            break;

    case 3: *sx = x-1;
            *sy = y-1;
            break;

    case 4: *sx = x-1;
            *sy = y;
            break;

    case 5: *sx = x-1;
            *sy = y+1;
            break;

    case 6: *sx = x;
            *sy = y+1;
            break;

    case 7: *sx = x+1;
            *sy = y+1;
            break;
  }
}

/* ----------------------------------------------------------------------- */
/*  Orientacin indica el pixel desde el que venimos                       */
/*  Devuelve la posicin relativa de (x,y) respecto a (sx,sy)              */

static int SiguientePixel (Imagen img, int orientacion, WORD *x, WORD *y)
{
  int dir;
  int xtmp;
  int ytmp;

  dir = (orientacion+1)%8;

  do {
      PixelAdyacente (*x,*y,dir,&xtmp,&ytmp);

      if (  xtmp<ANCHURA(img) && ytmp<ALTURA(img) && xtmp>=0 && ytmp>=0
         && PIXEL(img,xtmp,ytmp)) {
         *x = (WORD) xtmp;
         *y = (WORD) ytmp;
         return (dir+4)%8;
      }

      dir = (dir+1)%8;

  } while (dir!=orientacion);

  return -1;  /* No se puede seguir por este pixel */
}

/* ----------------------------------------------------------------------- */

int ExtraerContorno (Imagen img, int max, WORD *X, WORD *Y)
{
  WORD anchura = ANCHURA(img);
  WORD altura  = ALTURA(img);

  int  puntos  = 0;
  WORD x,y;
  int  orientacion;

  /* Punto inicial */

  for (x=0; x<anchura && !puntos; x++)
      for (y=0; y<altura && !puntos; y++)
          if (PIXEL(img,x,y)) {
             X[puntos] = x;
             Y[puntos] = y;
             puntos++;
          }

  /* Seguimiento del contorno */

  if (puntos) {

     x = X[0];
     y = Y[0];
     orientacion = 0;

     do {

         orientacion = SiguientePixel (img, orientacion, &x, &y);

         if ( orientacion!=-1 && !(x==X[0] && y==Y[0]) ) {
             X[puntos] = x;
             Y[puntos] = y;
             puntos++;
         }

     } while ( !(x==X[0] && y==Y[0]) && (puntos<max) && (orientacion!=-1) );
  }

  return puntos;
}

/* ----------------------------------------------------------------------- */

Imagen DibujarContorno (WORD anchura, WORD altura, int puntos, WORD *X, WORD *Y)
{
  int    i;
  Imagen tmp = CreaImagen (anchura, altura);

  for (i=0; i<puntos; i++)
      PIXEL(tmp,X[i],Y[i]) = 0xFF;

  return tmp;
}

/* ----------------------------------------------------------------------- */
/*  Gaussiana                                                              */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

#define MAX_GAUSS 64

double g[MAX_GAUSS];
double dg[MAX_GAUSS];
double d2g[MAX_GAUSS];

/* G(x) = exp((-x^2)/(2*(s^2)))                   */

static double gauss(double x, double sigma)
{
  /* return (double)exp((double) ((-x*x)/(2*sigma*sigma))); */

  return (1/(sigma*sqrt(2*M_PI)))*exp((double) ((-x*x)/(2*sigma*sigma)));
}

/* G'(x) = -x/(s^2)*gauss(x,s) = DoG */

static double dGauss (double x, double sigma)
{
  return - x / (sigma*sigma) * gauss(x, sigma);
}

/* G''(x) = -1/(s^2)*gauss(x,s) - x/(s^2)*dGauss(x,s) */

static double d2Gauss (double x, double sigma)
{
  return gauss(x,sigma)/(sigma*sigma)*((x*x)/(sigma*sigma)-1);
}


/* ----------------------------------------------------------------------- */
/*  Suavizamiento de un contorno                                           */
/* ----------------------------------------------------------------------- */
/*  s: Desviacin estndar de la gaussiana                                 */
/*  d: Tamao del filtro ---> Ventana 2d+1                                 */

void SuavizarContorno (float s, int d, int puntos, WORD *X, WORD *Y)
{
  int    i,k;
  int    p1,p2;
  double c, div;
  double tmp[MAX_CONTORNO];

  if (d>MAX_GAUSS)
     d = MAX_GAUSS;

  for (i=0; i<d; i++)
      g[i] = gauss ((double)i, s);

  div = g[0];

  for (i=1; i<d; i++)
      div += 2*g[i];

  /* X convolucionada con la gaussiana */

  for (i=0; i<puntos; i++) {

      c = g[0] * X[i];

      for (k=1; k<d; k++) {                  /* G(-x)=G(x) */
	  p1 = (i+k)%puntos;
	  p2 = (i-k+puntos)%puntos;
	  c += g[k]*X[p1] + g[k]*X[p2];
      }

      tmp[i] = c/div;
  }

  for (i=0; i<puntos; i++)
      X[i] = tmp[i];

  /* Y convolucionada con la gaussiana */

  for (i=0; i<puntos; i++) {

      c = g[0] * Y[i];

      for (k=1; k<d; k++) {                  /* G(-x)=G(x) */
	  p1 = (i+k)%puntos;
	  p2 = (i-k+puntos)%puntos;
	  c += g[k]*Y[p1] + g[k]*Y[p2];
      }

      tmp[i] = c/div;
  }

  for (i=0; i<puntos; i++)
      Y[i] = tmp[i];
}

/* ----------------------------------------------------------------------- */
/*  Curvatura de un contorno (Mokhtarian et al. 1986)                      */
/* ----------------------------------------------------------------------- */
/*  s: Desviacin estndar de la gaussiana                                 */
/*  d: Tamao del filtro ---> Ventana 2d+1                                 */

static void ConvolucionDG ( int d, int puntos, WORD *z, double *c )
{
  int i,k,p1,p2;

  for (i=0; i<puntos; i++) {

      c[i] = 0;                              /* G'(0) = 0 */

      for (k=1; k<d; k++) {                  /* G'(-x)=-G'(x) */
	  p1 = (i+k)%puntos;
	  p2 = (i-k+puntos)%puntos;
	  c[i] += -dg[k]*z[p1] + dg[k]*z[p2];
      }
  }
}

/* --- */

static void Convolucion2DG ( int d, int puntos, WORD *z, double *c )
{
  int i,k,p1,p2;

  for (i=0; i<puntos; i++) {

      c[i] = d2g[0]*z[i];

      for (k=1; k<d; k++) {                  /* G"(-x)=G"(x) */
	  p1 = (i+k)%puntos;
	  p2 = (i-k+puntos)%puntos;
	  c[i] += d2g[k]*z[p1] + d2g[k]*z[p2];
      }
  }
}

/* K = (x'y"-x"y')/((x'*x'+y'*y')^(3/2)) */

void CurvaturaContorno (float s, int d, int puntos, WORD *X, WORD *Y, double *K)
{
  int    i;
  double v,dx,dy;

  double x1[MAX_CONTORNO];
  double x2[MAX_CONTORNO];
  double y1[MAX_CONTORNO];
  double y2[MAX_CONTORNO];

  if (d>MAX_GAUSS)
     d = MAX_GAUSS;

  for (i=0; i<d; i++) {
      dg[i]  = dGauss  ((double)i, s);
      d2g[i] = d2Gauss ((double)i, s);
  }

  ConvolucionDG (d,puntos,X,x1);
  ConvolucionDG (d,puntos,Y,y1);

  Convolucion2DG (d,puntos,X,x2);
  Convolucion2DG (d,puntos,Y,y2);

  for (i=0; i<puntos; i++) {

      dx  = x1[i];
      dy  = y1[i];

      v   = dx*dx+dy*dy;

      if (v) {
         K[i] = (dx*y2[i]-x2[i]*dy)/sqrt(v*v*v);
      } else {
         K[i] = 1000000;  // Curvatura elevada (divisin por 0)
      }

  }
}

/* ----------------------------------------------------------------------- */

int PuntosRelevantesContorno ( float s, int d,
                               int puntos, WORD *X, WORD *Y,
                               WORD *Xmax, WORD *Ymax, WORD *Tmax)
{
  int    i,j;
  int    relevantes = 0;
  double K[MAX_CONTORNO];
  double max;

  CurvaturaContorno (s,d,puntos,X,Y,K);

  for (i=0; i<puntos; i++) {

      /* Comprueba si (X[i],Y[i]) es un punto relevante del contorno         */
      /* Puntos relevantes: mximos locales de curvatura en la ventana 2*d+1 */

      max = 0;

      for (j=-d; j<=d; j++)
          if ( j && abs(K[(i+j+puntos)%puntos])>max )
             max = abs(K[(i+j+puntos)%puntos]);

      if (max<=abs(K[i])) {
         /* Aade (X[i],Y[i]) a la lista de puntos relevantes del contorno */
         if (Xmax) Xmax[relevantes] = X[i];
         if (Ymax) Ymax[relevantes] = Y[i];
         if (Tmax) Tmax[relevantes] = (WORD)i;
         relevantes++;
      }
  }

  return relevantes;
}

/* ----------------------------------------------------------------------- */
/*  Linea: Dibuja una lnea segn el algoritmo de Bresenham                */
/* ----------------------------------------------------------------------- */

static void SwapInt( WORD *i1, WORD *i2 )
{
 WORD tmp = *i2;
 *i2 = *i1;
 *i1 = tmp;
}

static void Linea ( Imagen img, WORD x1, WORD y1, WORD x2, WORD y2, BYTE color)
{
  int d, dx, dy, aincr, bincr, xincr, yincr, x, y;

  if ( abs(x2-x1) < abs(y2-y1) ) {

     if ( y1 > y2 ) {
	SwapInt( &x1, &x2 );
	SwapInt( &y1, &y2 );
     }

     if ( x2 > x1 )
	xincr = 1;
     else
	xincr = -1;

     dy    = y2 - y1;
     dx    = abs( x2-x1 );
     d     = 2 * dx - dy;
     aincr = 2 * (dx - dy);
     bincr = 2 * dx;
     x     = x1;
     y     = y1;

     PIXEL(img,x,y) = color;

     for (y=y1+1; y<=y2; y++) {
	 if ( d >= 0 ) {
	    x += xincr;
	    d += aincr;
	 } else
	    d += bincr;

	 PIXEL(img,x,y) = color;
     }

  } else {

     if ( x1 > x2 ) {
	SwapInt( &x1, &x2 );
	SwapInt( &y1, &y2 );
     }

     if ( y2 > y1 )
	yincr = 1;
     else
	yincr = -1;

     dx    = x2 - x1;
     dy    = abs( y2-y1 );
     d     = 2 * dy - dx;
     aincr = 2 * (dy - dx);
     bincr = 2 * dy;
     x     = x1;
     y     = y1;

     PIXEL(img,x,y) = color;

     for (x=x1+1; x<=x2; x++) {
	 if ( d >= 0 ) {
	    y += yincr;
	    d += aincr;
	 } else
	    d += bincr;

	 PIXEL (img,x,y) = color;
     }
  }
}

/* --- */

Imagen AproximacionPoligonal ( WORD anchura, WORD altura,
                               float s, int d,
                               int puntos, WORD *X, WORD *Y )
{
  int    i,max;
  Imagen img = CreaImagen (anchura,altura);
  WORD   Xmax[MAX_CONTORNO];
  WORD   Ymax[MAX_CONTORNO];

  max = PuntosRelevantesContorno(s,d,puntos,X,Y,Xmax,Ymax,NULL);

  for (i=1; i<max; i++)
      Linea (img,Xmax[i-1],Ymax[i-1],Xmax[i],Ymax[i],0xFF);

  Linea (img,Xmax[max-1],Ymax[max-1],Xmax[0],Ymax[0],0xFF);

  return img;
}

/* ----------------------------------------------------------------------- */

