/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */
/*                      Frequency domain library                           */
/* ----------------------------------------------------------------------- */

#include "IMG_FREQ.H"

#include <stdlib.h>
#include <math.h>

/* ----------------------------------------------------------------------- */
/*  Nmeros complejos                                                      */
/* ----------------------------------------------------------------------- *

static Complex csum (Complex c1, Complex c2)    // c1+c2
{
  Complex sum;

  sum.real = c1.real + c2.real;
  sum.imag = c1.imag + c2.imag;

  return sum;
}

/* --- *

static Complex cdiff (Complex c1, Complex c2)    // c1-c2
{
  Complex diff;

  diff.real = c1.real - c2.real;
  diff.imag = c1.imag - c2.imag;

  return diff;
}

/* --- */

static Complex cprod (Complex c1, Complex c2)    // c1*c2
{
  Complex prod;

  prod.real = c1.real*c2.real - c1.imag*c2.imag;
  prod.imag = c1.real*c2.imag + c2.real*c1.imag;

  return prod;
}

/* --- *

static Complex cdiv (Complex c1, Complex c2)     // c1/c2
{
  Complex div;
  float   z;

  z = (c2.real)*(c2.real) + (c2.imag)*(c2.imag);

  if (z != 0.0) {
     div.real = (c1.real*c2.real + c1.imag*c2.imag)/z;
     div.imag = (c2.real*c1.imag - c1.real*c2.imag)/z;
  } else {
     div.real = 0.0;
     div.imag = 0.0;
  }

  return div;
}

/* --- *

static Complex cpot (Complex c, int n)     // c^n (n>0)
{
  Complex cn;
  int     i;

  cn.real = c.real;
  cn.imag = c.imag;

  for (i=2; i<=n; i++)
      cn = cprod(cn,c);

  return cn;
}

/* --- *

static float cnorm (Complex c)   // |c|^2
{
  return c.real*c.real + c.imag*c.imag;
}


/* --- */

static float cmod (Complex c)
{
  float norm = c.real*c.real + c.imag*c.imag;

  if (norm>0)
     return sqrt(norm);
  else
     return 0;
}

/* --- */

static float modulo (float cr, float ci)
{
  float norm = cr*cr + ci*ci;

  if (norm>0)
     return sqrt(norm);
  else
     return 0;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  FFT                                                                    */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

double  *sintab  = NULL;            /* Tabla de senos                      */
int     *bittab  = NULL;            /* Permutacin                         */
int     *powers  = NULL;            /* Potencias de 2                      */
int      puntos;                    /* Puntos de la FFT                    */
int      nn;                        /* 2^nn = puntos                       */
float    direction;                 /* Sentido de la FFT                   */
double   scale_factor = 0.0;        /* Factor 1/sqrt(puntos)               */

int      FFTN         = 0;          /* Tamao de las tablas de la FFT      */


#define FORWARD -1.0    /* FFT         */
#define REVERSE  1.0    /* FFT Inversa */


/* ----- */
/*  FFT  */
/* ----- */

/* datos - Array de datos            */
/* nivel - Nivel de recursin        */
/* chunk - Nodo inicial de la subFFT */

void FFFT ( Complex datos[], int nivel, int chunk )
{
  int     nodos;               /* Nmero de mariposas de la sub-FFT */
  int     i, sinindx, cosindx;
  int     dual1, dual2;                     /* Nodos de la mariposa */
  Complex dual1val, dual2val, dualprod, wp;

  nodos   = powers[nn-nivel];
  sinindx = bittab[chunk/nodos];     
  cosindx = (sinindx + puntos/4) % puntos;

  wp.real = sintab[cosindx];              /* W^p = exp(-j*2*PI*p/N) */
  wp.imag = direction * sintab[sinindx];

  for ( i=0; i<nodos; i++)  {  
      dual1         = chunk + i;
      dual2         = dual1 + nodos;
      dual1val.real = datos[dual1].real;
      dual1val.imag = datos[dual1].imag;
      dual2val.real = datos[dual2].real;
      dual2val.imag = datos[dual2].imag;
      dualprod.real = dual2val.real * wp.real - dual2val.imag * wp.imag;
      dualprod.imag = dual2val.real * wp.imag + dual2val.imag * wp.real;

      datos[dual1].real = dual1val.real + dualprod.real;
      datos[dual1].imag = dual1val.imag + dualprod.imag;
      datos[dual2].real = dual1val.real - dualprod.real;
      datos[dual2].imag = dual1val.imag - dualprod.imag;
  }

  if (nivel<nn)  {
     FFFT( datos, nivel + 1, chunk );
     FFFT( datos, nivel + 1, chunk + nodos);
  }
}

/* --- */

void fft ( Complex *datos, float dir )
{
  Complex tmpval;
  int     i;

  direction = dir;

  if ( direction != REVERSE )
     direction = FORWARD;

  FFFT( datos, 1, 0 );  /* FFT Recursiva: nivel 1, inicio en el nodo 0 */

  /* Reordenacin de valores */

  for ( i=0; i<puntos; i++ )
      if ( bittab[i]<=i )  {
	 tmpval.real           = datos[i].real;
	 tmpval.imag           = datos[i].imag;
	 datos[i].real         = datos[bittab[i]].real;
	 datos[i].imag         = datos[bittab[i]].imag;
	 datos[bittab[i]].real = tmpval.real;
	 datos[bittab[i]].imag = tmpval.imag;
      }

  /* Factor 1/sqrt(puntos) */

  for ( i=0; i<puntos; i++ )  {
      datos[i].real *= scale_factor;
      datos[i].imag *= scale_factor;
  }
}

/* ------------------------------------------------------------------------ */
/*  fftinit -- Inicializa las tablas usadas por la FFT                      */
/* ------------------------------------------------------------------------ */

static int bitrev( int bits )   /* Invertir bits */
{
  int i;
  int lookmask = 1;
  int setmask  = puntos>>1;
  int tmpbit   = 0;

  for ( i=0; i<=nn-1; i++ )  {  

      if ( bits & lookmask )
	 tmpbit = tmpbit | setmask;

      lookmask <<= 1;
      setmask  >>= 1;
  }

  return ( tmpbit );
}

/* --- */

void fftinit ( int num_datos )
{
  int  i;

  puntos       = num_datos;
  nn           = (log(puntos)/log(2.0)) + 0.5;   /* 2^nn = puntos */
  scale_factor = 1.0/sqrt(puntos);


  if ( sintab!=NULL )
     free ( sintab );

  if ( bittab!=NULL )
     free( bittab );

  if ( powers!=NULL )
     free( powers );


  sintab = (double*) calloc( puntos, sizeof(double) );
  bittab = (int*)    calloc( puntos, sizeof(int) );
  powers = (int*)    calloc( nn+1,   sizeof(int) );


  for ( i=0; i<=nn; i++ )
      powers[i] = pow( 2.0, i ) + .5;


  for ( i=0; i<puntos; i++ )  {
      sintab[i] = sin( 6.283185307179587 * i / puntos );
      bittab[i] = bitrev( i );
  }
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  fft2d -- Transformada Rpida de Fourier 2D de un array FFTN x FFTN.    */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

void fft2d ( Complex **imagen, float dir )
{
  Complex tmp[1024];   /* Para colocar la filas */
  int     i, j;        

  /* Transformada por columnas */

  for ( i=0; i<FFTN; i++ )
      fft (imagen[i], dir);

  /* Transformada por filas */

  for ( i=0; i<FFTN; i++ ) {

      for ( j=0; j<FFTN; j++ )  {
	  tmp[j].real = imagen[j][i].real;
	  tmp[j].imag = imagen[j][i].imag;
      }

      fft (tmp, dir);

      for ( j=0; j<FFTN; j++ ) {
	  imagen[j][i].real = tmp[j].real;
	  imagen[j][i].imag = tmp[j].imag;
      }
  }
}

/*  Modifica la entrada de forma que el origen (f=0) */
/* quede en el centro del array 2D.                  */

void filt_orig ( Complex **array )
{
  int  i, j;

  for ( i=0; i<FFTN; i++ )
      for ( j=0; j<FFTN; j++ )
	  if ( (i+j)%2 )  {
	     array[i][j].real = -array[i][j].real;
	     array[i][j].imag = -array[i][j].imag;
	  }

}

/* 2D FFT ( imagen 2^n x 2^n ) */

ComplexImg fftIMG (Imagen img, int oc)
{
  int        n;
  ComplexImg cim;

  if ( ANCHURA(img) == ALTURA(img) ) {

     n = ANCHURA(img);

     if (FFTN != n)
	fftinit (n);

     FFTN = n;

     cim = CreaComplexImg (img);

     if (oc)
	filt_orig (cim->datos);

     fft2d (cim->datos, FORWARD);

     return cim;

  } else

     return NULL;
}

/* 2D FFT inversa ( imagen 2^n x 2^n ) */

ComplexImg fftinvIMG (ComplexImg image, int oc)
{
  int        n;
  ComplexImg cim;

  n   = FFTN;
  cim = CopiaComplexImg ( image );

  if (FFTN != n) {
     fftinit (n);
     FFTN = n;
  }

  if (oc)
     filt_orig (cim->datos);

  fft2d (cim->datos, REVERSE);

  return cim;
}

/* --- */

ComplexImg FFTImagen (Imagen img)
{
  return fftIMG (img, 0);
}

ComplexImg OriginCenteredFFTImagen (Imagen img)
{
  return fftIMG (img, 1);
}


ComplexImg InverseFFTImagen (ComplexImg img)
{
  return fftinvIMG (img, 0);
}

ComplexImg InverseOriginCenteredFFTImagen (ComplexImg img)
{
  return fftinvIMG (img, 1);
}

/* ----------------------------------------------------------------------- */
/*  Adaptar imagen para que se le pueda aplicar la FFT (max 1024x1024)     */
/*  Se obtiene una imagen cuadrada de lado potencia de 2                   */

Imagen AdaptarImagenFFT (Imagen img, int dim)
{
  int    i,j,mr,mc;
  WORD   n;
  WORD   NR = ANCHURA(img);
  WORD   NC = ALTURA(img);
  Imagen salida;

  if (NR > NC)
     n = NR;
  else
     n = NC;

  if (dim<n) {
     if (n>512) {
	n = 1024;
     } else if (n > 256) {
	n = 512;
     } else if (n > 128) {
	n = 256;
     } else if (n > 64) {
	n = 128;
     } else if (n > 32) {
	n = 64;
     } else {
	n = 32;
     }
  } else
     n = dim;

  salida = CreaImagen (n,n);

  for (i=0; i<n; i++)
    for (j=0; j<n; j++)
	PIXEL(salida,i,j) = 0x00;

  mr = (n-NR)/2;
  mc = (n-NC)/2;

  for (i=mr; i<mr+NR; i++)
      for (j=mc; j<mc+NC; j++)
	  PIXEL(salida,i,j) = PIXEL(img,i-mr,j-mc);

  return salida;
}

/* --- */

Imagen ExtraerImagenFFT (Imagen img, WORD anchura, WORD altura)
{
  int    i,j,mr,mc;
  WORD   n;
  Imagen salida;

  if (anchura > altura)
     n = anchura;
  else
     n = altura;

  if (n>512) {
     n = 1024;
  } else if (n > 256) {
     n = 512;
  } else if (n > 128) {
     n = 256;
  } else if (n > 64) {
     n = 128;
  } else if (n > 32) {
     n = 64;
  } else {
     n = 32;
  }

  salida = CreaImagen (anchura,altura);

  mr = (n-anchura)/2;
  mc = (n-altura)/2;

  for (i=mr; i<mr+anchura; i++)
      for (j=mc; j<mc+altura; j++)
	  PIXEL(salida,i-mr,j-mc) = PIXEL(img,i,j);

  return salida;
}


/* ----------------------------------------------------------------------- */
/*  Devuelve una imagen de la FFT de la imagen 'img'                       */

Imagen ImagenFFT (Imagen img)
{
  Imagen     aux    = AdaptarImagenFFT (img,0);
  ComplexImg fft    = FFTImagen (aux);
  Imagen     salida = ImagenComplex (fft, fft->anchura, fft->altura);

  DestruyeComplexImg (&fft);
  DestruyeImagen (&aux);
  return salida;
}

Imagen ImagenLogFFT (Imagen img)
{
  Imagen     aux    = AdaptarImagenFFT (img,0);
  ComplexImg fft    = FFTImagen (aux);
  Imagen     salida = ImagenLogComplex (fft, fft->anchura, fft->altura);

  DestruyeComplexImg (&fft);
  DestruyeImagen (&aux);
  return salida;
}


/* ----------------------------------------------------------------------- */
/* Devuelve una imagen de la FFT (origin-centered) de la imagen 'img'      */

Imagen ImagenOriginCenteredFFT (Imagen img)
{
  Imagen     aux    = AdaptarImagenFFT (img,0);
  ComplexImg fft    = OriginCenteredFFTImagen (aux);
  Imagen     salida = ImagenComplex (fft, fft->anchura, fft->altura);

  DestruyeComplexImg (&fft);
  DestruyeImagen (&aux);
  return salida;
}

Imagen ImagenLogOriginCenteredFFT (Imagen img)
{
  Imagen     aux    = AdaptarImagenFFT (img,0);
  ComplexImg fft    = OriginCenteredFFTImagen (aux);
  Imagen     salida = ImagenLogComplex (fft, fft->anchura, fft->altura);

  DestruyeComplexImg (&fft);
  DestruyeImagen (&aux);
  return salida;
}

/* ----------------------------------------------------------------------- */
/* Reordena una imagen (desordenada por la FFT)                            */

Imagen Shuffle (Imagen img)
{
  int    n,k;
  Imagen p;
  int    i,j;

  n = ANCHURA(img); // == ALTURA(img)
  k = n/2;
  p = CreaImagen (n, n);

  for (i=0; i<k; i++)
      for (j=0; j<k; j++) {
	  PIXEL(p,i,j)     = PIXEL(img,i+k,j+k);
	  PIXEL(p,i+k,j+k) = PIXEL(img,i,j);
	  PIXEL(p,i,j+k)   = PIXEL(img,i+k,j);
	  PIXEL(p,i+k,j)   = PIXEL(img,i,j+k);
      }

  return p;
}

/* ----------------------------------------------------------------------- */
/*  Filtro ideal paso bajo                                                 */

Imagen LowPassFilter (Imagen img, int r)
{
  int        dim;
  WORD       anchura = ANCHURA(img);
  WORD       altura  = ALTURA(img);
  Imagen     aux,salida;
  ComplexImg fft1, fft2;
  int        i,j, hc, vc, dist;

  /* Transformada */

  aux = AdaptarImagenFFT(img,0);
  dim = ANCHURA(aux);

  fft1 = OriginCenteredFFTImagen (aux);

  /* Filtrado */

  hc = dim/2;
  vc = dim/2;

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++) {
	  dist = modulo(i-vc,j-hc);
	  if (dist > r) {
	     fft1->datos[i][j].real = 0.0;
	     fft1->datos[i][j].imag = 0.0;
	  }
      }

  /* Transformada inversa */

  fft2   = InverseFFTImagen (fft1);
  salida = ImagenComplex (fft2,anchura,altura);

  DestruyeComplexImg(&fft1);
  DestruyeComplexImg(&fft2);
  DestruyeImagen (&aux);

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  Filtro ideal paso alto                                                 */

Imagen HighPassFilter (Imagen img, int r)
{
  int        dim;
  WORD       anchura = ANCHURA(img);
  WORD       altura  = ALTURA(img);
  Imagen     aux,salida;
  ComplexImg fft1, fft2;
  int        i,j, hc, vc, dist;

  /* Transformada */

  aux = AdaptarImagenFFT(img,0);
  dim = ANCHURA(aux);

  fft1 = OriginCenteredFFTImagen (aux);

  /* Filtrado */

  hc = dim/2;
  vc = dim/2;

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++) {
	  dist = modulo(i-vc,j-hc);
	  if (dist <= r) {
	     fft1->datos[i][j].real = 0.0;
	     fft1->datos[i][j].imag = 0.0;
	  }
      }

  /* Transformada inversa */

  fft2   = InverseFFTImagen (fft1);
  salida = ImagenComplex (fft2,anchura,altura);

  DestruyeComplexImg(&fft1);
  DestruyeComplexImg(&fft2);
  DestruyeImagen(&aux);

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  Filtro ideal paso banda                                                */

Imagen BandPassFilter (Imagen img, int r1, int r2)
{
  int        dim;
  WORD       anchura = ANCHURA(img);
  WORD       altura  = ALTURA(img);
  Imagen     aux,salida;
  ComplexImg fft1, fft2;
  int        i,j, hc, vc, dist;

  /* Transformada */

  aux = AdaptarImagenFFT(img,0);
  dim = ANCHURA(aux);

  fft1 = OriginCenteredFFTImagen (aux);

  /* Filtrado */

  hc = dim/2;
  vc = dim/2;

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++) {
	  dist = modulo(i-vc,j-hc);
	  if ((dist<=r1) || (dist>r2)) {
	     fft1->datos[i][j].real = 0.0;
	     fft1->datos[i][j].imag = 0.0;
	  }
      }

  /* Transformada inversa */

  fft2   = InverseFFTImagen (fft1);
  salida = ImagenComplex (fft2,anchura,altura);

  DestruyeComplexImg(&fft1);
  DestruyeComplexImg(&fft2);
  DestruyeImagen(&aux);

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  Filtro Butterworth paso bajo                                           */

Imagen ButterworthLowPass (Imagen img, int d, int n)
{
  int        dim;
  WORD       anchura = ANCHURA(img);
  WORD       altura  = ALTURA(img);
  Imagen     aux,salida;
  ComplexImg fft1, fft2;
  int        i,j, hc, vc;
  double     dist,H;

  /* Transformada */

  aux = AdaptarImagenFFT(img,0);
  dim = ANCHURA(aux);

  fft1 = OriginCenteredFFTImagen (aux);

  /* Filtrado */

  hc = dim/2;
  vc = dim/2;

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++) {
	  dist = modulo(i-vc,j-hc);
	  H    = pow((double)dist/d,(double)2.0*n);
	  H    = 1 / (1+H);
	  fft1->datos[i][j].real *= H;
          fft1->datos[i][j].imag *= H;
      }

  /* Transformada inversa */

  fft2   = InverseFFTImagen (fft1);
  salida = ImagenComplex (fft2,anchura,altura);

  DestruyeComplexImg(&fft1);
  DestruyeComplexImg(&fft2);
  DestruyeImagen (&aux);

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  Filtro Butterworth paso alto                                           */

Imagen ButterworthHighPass (Imagen img, int d, int n)
{
  int        dim;
  WORD       anchura = ANCHURA(img);
  WORD       altura  = ALTURA(img);
  Imagen     aux,salida;
  ComplexImg fft1, fft2;
  int        i,j, hc, vc;
  double     dist,H;

  /* Transformada */

  aux = AdaptarImagenFFT(img,0);
  dim = ANCHURA(aux);

  fft1 = OriginCenteredFFTImagen (aux);

  /* Filtrado */

  hc = dim/2;
  vc = dim/2;

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++) {
	  dist = modulo(i-vc,j-hc);

          if (dist>0) {
	      H = pow(d/dist,2*n);
	      H = 1 / (1+H);
	  } else
              H = 0;

	  fft1->datos[i][j].real *= H;
          fft1->datos[i][j].imag *= H;
      }

  /* Transformada inversa */

  fft2   = InverseFFTImagen (fft1);
  salida = ImagenComplex (fft2,anchura,altura);

  DestruyeComplexImg(&fft1);
  DestruyeComplexImg(&fft2);
  DestruyeImagen(&aux);

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  Filtro exponencial paso bajo                                           */

Imagen ExponentialLowPass (Imagen img, int sup)
{
  int        dim;
  WORD       anchura = ANCHURA(img);
  WORD       altura  = ALTURA(img);
  Imagen     aux,salida;
  ComplexImg fft1, fft2;
  int        i,j, hc, vc;
  float      dist,H;

  /* Transformada */

  aux = AdaptarImagenFFT(img,0);
  dim = ANCHURA(aux);

  fft1 = OriginCenteredFFTImagen (aux);

  /* Filtrado */

  hc = dim/2;
  vc = dim/2;

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++) {
	  dist = modulo(i-vc,j-hc);

          H = exp(-0.347*(dist/sup));

	  fft1->datos[i][j].real *= H;
          fft1->datos[i][j].imag *= H;
      }

  /* Transformada inversa */

  fft2   = InverseFFTImagen (fft1);
  salida = ImagenComplex (fft2,anchura,altura);

  DestruyeComplexImg(&fft1);
  DestruyeComplexImg(&fft2);
  DestruyeImagen(&aux);

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  Filtro exponencial paso alto                                           */

Imagen ExponentialHighPass (Imagen img, int inf)
{
  int        dim;
  WORD       anchura = ANCHURA(img);
  WORD       altura  = ALTURA(img);
  Imagen     aux,salida;
  ComplexImg fft1, fft2;
  int        i,j, hc, vc;
  float      dist,H;

  /* Transformada */

  aux = AdaptarImagenFFT(img,0);
  dim = ANCHURA(aux);

  fft1 = OriginCenteredFFTImagen (aux);

  /* Filtrado */

  hc = dim/2;
  vc = dim/2;

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++) {
	  dist = modulo(i-vc,j-hc);

          if (dist>0)
             H = exp(-0.347*(inf/dist));
          else
             H = 0;

	  fft1->datos[i][j].real *= H;
          fft1->datos[i][j].imag *= H;
      }

  /* Transformada inversa */

  fft2   = InverseFFTImagen (fft1);
  salida = ImagenComplex (fft2,anchura,altura);

  DestruyeComplexImg(&fft1);
  DestruyeComplexImg(&fft2);
  DestruyeImagen(&aux);

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  Filtro exponencial paso banda (sup>inf)                                */

Imagen ExponentialBandPass (Imagen img, int inf, int sup)
{
  int        dim;
  WORD       anchura = ANCHURA(img);
  WORD       altura  = ALTURA(img);
  Imagen     aux,salida;
  ComplexImg fft1, fft2;
  int        i,j, hc, vc;
  float      dist,H;

  float      prod = (float)inf*(float)sup;
  float      diff = (float)sup-(float)inf;

  /* Transformada */

  aux = AdaptarImagenFFT(img,0);
  dim = ANCHURA(aux);

  fft1 = OriginCenteredFFTImagen (aux);

  /* Filtrado */

  hc = dim/2;
  vc = dim/2;

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++) {
	  dist = modulo(i-vc,j-hc);

          if (dist>0)
             H = exp(-0.347*((prod-dist*dist)/(diff*dist)));
          else
             H = 0;

	  fft1->datos[i][j].real *= H;
          fft1->datos[i][j].imag *= H;
      }

  /* Transformada inversa */

  fft2   = InverseFFTImagen (fft1);
  salida = ImagenComplex (fft2,anchura,altura);

  DestruyeComplexImg(&fft1);
  DestruyeComplexImg(&fft2);
  DestruyeImagen(&aux);

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  Cross-Section filter                                                   */


/* Ejemplo: High emphasis filter


  float d[] = { 0.5, 0.55, 0.6, 0.8, 1.0, 1.3, 1.5, 1.6, 1.7,
	      1.8, 1.85, 1.9, 1.94, 1.97, 1.99, 2.0, 2.0 };

  Nd = 16


  float d[] = { 0.5,  0.51, 0.53, 0.55, 0.6, 0.8, 1.0, 1.3, 1.5, 1.6, 1.7,
	      1.8, 1.85, 1.9, 1.94, 1.97, 1.99, 2.0, 2.0};

  Nd = 18

 */ 

Imagen CrossSectionFilter (Imagen img, float *d, int Nd)
{
  int        dim;
  WORD       anchura = ANCHURA(img);
  WORD       altura  = ALTURA(img);
  Imagen     aux,salida;
  ComplexImg fft, inv_fft;
  int        i,j, hc, vc, dist, ia, ib;
  float      a, b;

  int        FMAX = Nd-1;


  aux = AdaptarImagenFFT(img,0);
  dim = ANCHURA(aux);

  fft = OriginCenteredFFTImagen (aux);

  DestruyeImagen (&aux);

  hc = dim/2;
  vc = dim/2;

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++) {

	   dist = modulo(i-vc,j-hc);

	   if (dist >= FMAX) {
	      fft->datos[i][j].real *= d[FMAX];
              fft->datos[i][j].imag *= d[FMAX];
	   } else if (dist > 0) {
	      ia = dist;
	      ib = ia+1;
	      a  = d[ia];
	      b  = d[ib];
	      fft->datos[i][j].real *= (float)(dist-ia)*a + (float)(ib-dist)*b;
              fft->datos[i][j].imag *= (float)(dist-ia)*a + (float)(ib-dist)*b;
	   } else {
	      fft->datos[i][j].real *= d[0];
	      fft->datos[i][j].imag *= d[0];
           }
      }

  inv_fft = InverseFFTImagen (fft);
  salida  = ImagenComplex (inv_fft,anchura,altura);

  DestruyeComplexImg(&fft);
  DestruyeComplexImg(&inv_fft);

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  Homomorphic filter                                                     */

Imagen HomomorphicFilter (Imagen img, float *d, int Nd)
{
  int        dim;
  WORD       anchura = ANCHURA(img);
  WORD       altura  = ALTURA(img);
  Imagen     aux,salida;
  ComplexImg fft, inv_fft;
  int        i,j, hc, vc, ia, ib, dist;
  float      a, b;

  int FMAX = Nd - 1;


  aux = AdaptarImagenFFT(img,0);
  dim = ANCHURA(aux);

  if (FFTN != dim) {
     fftinit (dim);
     FFTN = dim;
  }

  /* Logaritmo de la imagen */

  fft = CreaComplexImg(aux);

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++) {
	  fft->datos[i][j].real = log ( PIXEL(aux,i,j)+1 );
	  fft->datos[i][j].imag = 0.0;
      }

  DestruyeImagen(&aux);

  /* Transformada de Fourier */

  filt_orig (fft->datos);
  fft2d     (fft->datos,FORWARD);

  /* Filtrado */

  vc = dim/2;
  hc = dim/2;

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++) {

	   dist = modulo(i-vc,j-hc);

	   if (dist >= FMAX) {
	      fft->datos[i][j].real *= d[FMAX];
              fft->datos[i][j].imag *= d[FMAX];
	   } else if (dist > 0) {
	      ia = dist;
	      ib = ia+1;
	      a  = d[ia];
	      b  = d[ib];
	      fft->datos[i][j].real *= (float)(dist-ia)*a + (float)(ib-dist)*b;
              fft->datos[i][j].imag *= (float)(dist-ia)*a + (float)(ib-dist)*b;
	   } else {
	      fft->datos[i][j].real *= d[0];
	      fft->datos[i][j].imag *= d[0];
           }
      }

  /* Transformada inversa */

  inv_fft = InverseFFTImagen (fft);

  DestruyeComplexImg (&fft);

  /* Exponencial */

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++) {
	  a = cmod (inv_fft->datos[i][j]);
	  inv_fft->datos[i][j].real = (float) exp((double)a) - 1.0;
          inv_fft->datos[i][j].imag = 0;
      }

  salida = ImagenComplex(inv_fft,anchura,altura);

  DestruyeComplexImg(&inv_fft);

  return salida;
}

/* ----------------------------------------------------------------------- */

Imagen FourierConvolve ( Imagen   img,
                         FloatImg mask, int dimX, int dimY, int centroX, int centroY)
{
  int        dim;
  WORD       anchura = ANCHURA(img);
  WORD       altura  = ALTURA(img);
  Imagen     aux,salida;
  ComplexImg fft_img, fft_inv, fft_mask;
  int        i,j,x,y,hc, vc, dist;

  /* Transformada de la imagen */

  aux = AdaptarImagenFFT(img,0);
  dim = ANCHURA(aux);

  fft_img = OriginCenteredFFTImagen (aux);

  DestruyeImagen (&aux);

  /* Transformada de la mscara */

  fft_mask = CreaComplex(dim,dim);

  for (i=0; i<dimX; i++)
      for (j=0; j<dimY; j++) {
          x = (i-centroX+dim)%dim;
          y = (j-centroY+dim)%dim;
          fft_mask->datos[x][y].real = mask->datos[i][j];
      }

  filt_orig (fft_mask->datos);
  fft2d (fft_mask->datos, FORWARD);

  /* Filtrado */

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++)
          fft_img->datos[i][j] = cprod (fft_img->datos[i][j], fft_mask->datos[i][j]);

  /* Transformada inversa */

  fft_inv = InverseFFTImagen (fft_img);
  salida  = ImagenComplex (fft_inv,anchura,altura);

  DestruyeComplexImg(&fft_img);
  DestruyeComplexImg(&fft_inv);
  DestruyeComplexImg(&fft_mask);

  return salida;
}

/* -----------------------------------------------------------------------

Imagen FrequencyDomainFilter (Imagen img)
{
  Imagen     salida;
  ComplexImg fft1, fft2;

  if (ANCHURA(img)!=ALTURA(img))
     return NULL;

  fft1   = OriginCenteredFFTImagen (img);

  // Filter

  fft2   = InverseFFTImagen (fft1);
  salida = ImagenComplex (fft2);

  DestruyeComplexImg(fft1);
  DestruyeComplexImg(fft2);

  return salida;
}

  ----------------------------------------------------------------------- */
