/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */
/*                               Wavelets                                  */
/* ----------------------------------------------------------------------- */

#include "IMG_WAVE.H"

#include <stdlib.h>
#include <math.h>

/* ----------------------------------------------------------------------- */

#define FORWARD  1
#define INVERSE -1

#define MAX_N   1024

/* ----------------------------------------------------------------------- */

/* void wavelet (float *data, int N, int direction)                        */

float tmp[MAX_N];      /* Array auxiliar usado por las funciones 'wavelet' */

/* Haar basis */

void Haar_2 (float *data, int N, int direction)
{
  int    i,j;
  int    nover2 = N/2;
  float  h0,h1;

  h0 = h1 = 0.5;  /* h0 = h1 = 1.0/sqrt(2.0); */

  if (direction == FORWARD) {
     i=0;
     for (j=0; j<N-1; j+=2) {
	 tmp[i]        = h0*data[j] + h1*data[j+1];
	 tmp[i+nover2] = h0*data[j] - h1*data[j+1];
	 i++;
     }
  } else if (direction == INVERSE) {
     /* Comment the next line out for symmetric scaling: */
     h1 = h0 = 1.0;
     i = j = 0;
     do {
	 tmp[j]   = h0*data[i] + h1*data[i+nover2];
	 tmp[j+1] = h0*data[i] - h1*data[i+nover2];
	 j += 2;
	 i++;
     } while (i < nover2);
  }

  for (i=0; i<N; i++)
      data[i] = tmp[i];
}

/* Daubechies basis (4 coefficients) */

void Daub_4 (float *data, int N, int direction)
{
  int    i,j,nover2;

  static int   init = 0;
  static float h0,h1,h2,h3;

  if (init == 0) {
       double rt3, rt2;

       rt3 = sqrt(3.0);
       h0 = (1.0 + rt3)/4.0;
       h1 = (3.0 + rt3)/4.0;
       h2 = (3.0 - rt3)/4.0;
       h3 = (1.0 - rt3)/4.0;

       rt2 = sqrt(2.0);        /* Symmetrical scaling */

       h0 /= rt2;
       h1 /= rt2;
       h2 /= rt2;
       h3 /= rt2;
       init = 1;
  }

  nover2 = N/2;

  if (direction>=0) {

     i=0;
     for (j=0; j<N-3; j+=2) {
	 tmp[i] = h0*data[j] + h1*data[j+1] + h2*data[j+2] + h3*data[j+3];
	 tmp[i+nover2] = -(h0*data[j+3] + h2*data[j+1]) + (h1*data[j+2] + h3*data[j]);
	 i++;
     }

     /* End around overlap */

     tmp[i] = h0*data[N-2] + h1*data[N-1] + h2*data[0] + h3*data[1];
     tmp[i+nover2] = -(h0*data[1] + h2*data[N-1]) + (h1*data[0] + h3*data[N-2]);

  } else {

     i = 0;
     j = 2;

     do {
	 tmp[j] = h2*data[i]+h1*data[i+nover2]+
		  h0*data[(i+1)%nover2]+h3*data[(i+1)%nover2 + nover2];
	 tmp[j+1]=h3*data[i]-h0*data[i+nover2]+
		  h1*data[(i+1)%nover2]-h2*data[(i+1)%nover2 + nover2];
	 j=(j+2)%N;
	 i++;
     } while (i < nover2);
  }

  for (i=0; i<N; i++)
      data[i] = tmp[i];

}

/* Fast Wavelet 1D: N potencia de dos */

void fwt_1d ( float *data, int N, int direction,
	      void (*wavelet)(float*,int,int), int coef )
{
  int thisn;

  if (N > coef) {
     if (direction >= 0) {
	thisn = N;
	while (thisn >= coef) {
	      wavelet (data, thisn, direction);
	      thisn /= 2;
	}
     } else if (direction == INVERSE) {
	thisn = coef;
	while (thisn <= N) {
	      wavelet (data, thisn, direction);
	      thisn *= 2;
	}
     }
  }
}

/* Fast Wavelet 2D: NxN. N potencia de dos */

void fwt_2d ( float **result, int N, int direction,
	      void (*wavelet)(float*,int,int), int coef )
{
  int   i,j;
  float tmp[MAX_N];

  for (i=0; i<N; i++)
      fwt_1d (result[i], N, direction, wavelet, coef);

  for (j=0; j<N; j++) {
      for (i=0; i<N; i++)
	  tmp[i] = result[i][j];
      fwt_1d (tmp, N, direction, wavelet, coef);
      for (i=0; i<N; i++)
	  result[i][j] = tmp[i];
  }
}

/* ----------------------------------------------------------------------- */

Wavelet CreaWavelet (int n)
{
  int     i;
  Wavelet wt = (Wavelet) malloc(sizeof(struct InfoWavelet));

  if (wt) {
     wt->info = (float **) calloc (sizeof(float*),n);
     wt->n    = n;

     for (i=0; i<n; i++)
	 wt->info[i] = (float *) calloc (sizeof(float),n);
  }

  return wt;
}


void DestruyeWavelet (Wavelet *wt)
{
  int i;

  if (*wt) {

     for (i=0; i<(*wt)->n; i++)
	 free((*wt)->info[i]);

     free ((*wt)->info);
     free ((*wt));
     *wt = NULL;
  }
}


Wavelet CopiaWavelet (Wavelet wt)
{
  int     n      = wt->n;
  Wavelet result = CreaWavelet(n);
  int     i, j;

  for (i=0; i<n; i++)
      for (j=0; j<n; j++)
	  result->info[i][j] = wt->info[i][j];

  return result;
}

/* ----------------------------------------------------------------------- */

void ReduceWavelet (Wavelet wt, int n)   /* Conserva slo NxN coeficientes */
{
  int i,j;

  if (n<(wt->n))
     for (i=n; i<(wt->n); i++)
	 for (j=n; j<(wt->n); j++)
	     wt->info[i][j] = 0.0;
}

/* ----------------------------------------------------------------------- */
/*  2D wavelet transform, using Haar basis                                 */
/* ----------------------------------------------------------------------- */

void Haar (Wavelet wt)
{
  fwt_2d (wt->info, wt->n, FORWARD, Haar_2, 2);
}
/* --- */

void InverseHaar (Wavelet wt)
{
  fwt_2d (wt->info, wt->n, INVERSE, Haar_2, 2);
}

/* ----------------------------------------------------------------------- */
/*  2D wavelet transform, using Daubechies basis (4 coefficients)          */
/* ----------------------------------------------------------------------- */

void Daubechies (Wavelet wt)
{
  fwt_2d (wt->info, wt->n, FORWARD, Daub_4, 4);
}

/* --- */

void InverseDaubechies (Wavelet wt)
{
  fwt_2d (wt->info, wt->n, INVERSE, Daub_4, 4);
}

/* ----------------------------------------------------------------------- */
/*  2D Walsh transform                                                     */
/* ----------------------------------------------------------------------- */

int logn (int n)
{
  int i,j;

  j = 1;
  i = 0;

  while (j<n) {
	j = j << 1;
	i++;
  }

  return i;
}


void Walsh1D (float *f, int n)
{
  int   i,j,k,k1,k2;
  float tmp;
  int   ln  = logn(n);

  j = 0;

  for (i=0; i<n-1; i++) {
      if (i < j) {
	 tmp = f[i];
	 f[i] = f[j];
	 f[j] = tmp;
      }
      k = n/2;
      while (k<j+1) {
	    j -= k;
	    k /= 2;
      }
      j += k;
  }

  for (k=1; k<=ln; k++) {
      k2 = 1<<k;
      k1 = 1<<(k-1);
      for (j=0; j<k1; j++) {
	  for (i=j; i<n; i+= k2) {
	      tmp = f[i+k1];
	      f[i+k1] = f[i] - tmp;
	      f[i]    = f[i] + tmp;
	  }
      }
  }

  for (i=0; i<n; i++)
      f[i] = f[i]/(float)n;
}


void Walsh2D (float **res, int n)
{
  long  i,j;
  float data[MAX_N];

  for (i=0; i<n; i++)  {
      for (j=0; j<n; j++)
	  data[j] = res[i][j];
      Walsh1D (data, n);
      for (j=0; j<n; j++)
	  res[i][j] = data[j];
  }

  for (j=0; j<n; j++)   {
      for (i=1; i<n; i++)
	  data[i] = res[i][j];
      Walsh1D(data, n);
      for (i=0; i<n; i++)
	  res[i][j] = data[i];
  }
}


void UnWalsh2D (float **res, int n)
{
  float data[MAX_N];
  int   i,j;

  for (i=0; i<n; i++) {
      for (j=0; j<n; j++)
	  data[j] = res[i][j];
      Walsh1D (data, n);
      for (j=0; j<n; j++)
	  res[i][j] = data[j]*n;
  }

  for (j=0; j<n; j++) {
      for (i=0; i<n; i++)
	  data[i] = res[i][j];
      Walsh1D(data, n);
      for (i=0; i<n; i++)
	  res[i][j] = data[i]*n;
  }
}

/* --- */

void Walsh (Wavelet wt)
{
  Walsh2D (wt->info, wt->n);
}

/* --- */

void InverseWalsh (Wavelet wt)
{
  UnWalsh2D (wt->info, wt->n);
}

/* ----------------------------------------------------------------------- */
/*  Imagen ---> Imagen cuadrada de lado potencia de 2 (max MAX_NxMAX_N)    */

Wavelet WaveletImagen (Imagen img)
{
  int     i,j,mr,mc;
  WORD    n;
  WORD    NR = ANCHURA(img);
  WORD    NC = ALTURA(img);
  Wavelet salida;

  if (NR > NC)
     n = NR;
  else
     n = NC;

  if (n>512) {
     n = 1024;
  } else if (n > 256) {
     n = 512;
  } else if (n > 128) {
     n = 256;
  } else if (n > 64) {
     n = 128;
  } else if (n > 32) {
     n = 64;
  } else {
     n = 32;
  }

  salida = CreaWavelet(n);

  for (i=0; i<n; i++)
    for (j=0; j<n; j++)
	salida->info[i][j] = 0;

  mr = (n-NR)/2;
  mc = (n-NC)/2;

  for (i=mr; i<mr+NR; i++)
      for (j=mc; j<mc+NC; j++)
	  salida->info[i][j] = PIXEL(img,i-mr,j-mc);

  return salida;
}

/* --- */

Imagen ImagenWavelet (Wavelet wt)
{
  Imagen img = CreaImagen(wt->n,wt->n);
  float  xmax,xmin,xd;
  int    i,j;

  xmax = wt->info[0][0];
  xmin = wt->info[0][0];

  for (i=0; i<(wt->n); i++)
      for (j=0; j<(wt->n); j++) {
	    if (xmax < wt->info[i][j]) xmax = wt->info[i][j];
	    if (xmin > wt->info[i][j]) xmin = wt->info[i][j];
      }

  xd = xmax-xmin;

  if (xd)
     for (i=0; i<(wt->n); i++)
         for (j=0; j<(wt->n); j++)
	     PIXEL(img,i,j) = (wt->info[i][j]-xmin)/xd * 255;

  return img;
}

/* --- */

Imagen ImagenLogWavelet (Wavelet wt)
{
  Imagen img = CreaImagen(wt->n,wt->n);
  float  xmax,xmin,xd;
  int    i,j;

  xmax = wt->info[0][0];
  xmin = wt->info[0][0];

  for (i=0; i<(wt->n); i++)
      for (j=0; j<(wt->n); j++) {
	    if (xmax < wt->info[i][j]) xmax = wt->info[i][j];
	    if (xmin > wt->info[i][j]) xmin = wt->info[i][j];
      }

  xd = xmax-xmin;
  xd = log(1+xd);

  if (xd)
     for (i=0; i<(wt->n); i++)
         for (j=0; j<(wt->n); j++)
	     PIXEL(img,i,j) = log((wt->info[i][j]-xmin)+1)/xd * 255;

  return img;
}

/* --- */

Imagen ExtraerImagenWavelet (Wavelet wt, WORD anchura, WORD altura)
{
  Imagen img = CreaImagen(anchura,altura);
  float  xmax,xmin,xd;
  int    i,j,mr,mc;
  WORD   n = wt->n;

  mr = (n-anchura)/2;
  mc = (n-altura)/2;

  xmax = wt->info[mr][mr];
  xmin = wt->info[mr][mr];

  for (i=mr; i<mr+anchura; i++)
      for (j=mc; j<mc+altura; j++) {
	    if (xmax < wt->info[i][j]) xmax = wt->info[i][j];
	    if (xmin > wt->info[i][j]) xmin = wt->info[i][j];
      }

  xd = xmax-xmin;

  if (xd)
     for (i=mr; i<mr+anchura; i++)
         for (j=mc; j<mc+altura; j++)
	     PIXEL(img,i-mr,j-mc) = (wt->info[i][j]-xmin)/xd * 255;

  return img;
}


/* ----------------------------------------------------------------------- */
/*  Devuelve una imagen de la transformada de la imagen 'img'              */

Imagen ImagenHaar (Imagen img)
{
  Wavelet wt;
  Imagen  salida;

  wt = WaveletImagen (img);

  Haar (wt);

  salida = ImagenWavelet (wt);

  DestruyeWavelet (&wt);

  return salida;
}


Imagen ImagenLogHaar (Imagen img)
{
  Wavelet wt;
  Imagen  salida;

  wt = WaveletImagen (img);

  Haar (wt);

  salida = ImagenLogWavelet (wt);

  DestruyeWavelet (&wt);

  return salida;
}

/* --- */

Imagen ImagenDaubechies (Imagen img)
{
  Wavelet wt;
  Imagen  salida;

  wt = WaveletImagen (img);

  Daubechies (wt);

  salida = ImagenWavelet (wt);

  DestruyeWavelet (&wt);

  return salida;
}


Imagen ImagenLogDaubechies (Imagen img)
{
  Wavelet wt;
  Imagen  salida;

  wt = WaveletImagen (img);

  Daubechies (wt);

  salida = ImagenLogWavelet (wt);

  DestruyeWavelet (&wt);

  return salida;
}

/* --- */

Imagen ImagenWalsh (Imagen img)
{
  Wavelet wt;
  Imagen  salida;

  wt = WaveletImagen (img);

  Walsh (wt);

  salida = ImagenWavelet (wt);

  DestruyeWavelet (&wt);

  return salida;
}


Imagen ImagenLogWalsh (Imagen img)
{
  Wavelet wt;
  Imagen  salida;

  wt = WaveletImagen (img);

  Walsh (wt);

  salida = ImagenLogWavelet (wt);

  DestruyeWavelet (&wt);

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  "Filtrado" mediante una transformacin                                 */

Imagen FiltradoHaar (Imagen img, int n)
{
  Wavelet wt;
  Imagen  salida;

  wt = WaveletImagen (img);

  Haar (wt);
  ReduceWavelet(wt,n);
  InverseHaar(wt);

  salida = ExtraerImagenWavelet (wt, ANCHURA(img), ALTURA(img));

  DestruyeWavelet (&wt);

  return salida;
}

/* --- */

Imagen FiltradoDaubechies (Imagen img, int n)
{
  Wavelet wt;
  Imagen  salida;

  wt = WaveletImagen (img);

  Daubechies (wt);
  ReduceWavelet(wt,n);
  InverseDaubechies(wt);

  salida = ExtraerImagenWavelet (wt, ANCHURA(img), ALTURA(img));

  DestruyeWavelet (&wt);

  return salida;
}

/* --- */

Imagen FiltradoWalsh (Imagen img, int n)
{
  Wavelet wt;
  Imagen  salida;

  wt = WaveletImagen (img);

  Walsh (wt);
  ReduceWavelet(wt,n);
  InverseWalsh(wt);

  salida = ExtraerImagenWavelet (wt, ANCHURA(img), ALTURA(img));

  DestruyeWavelet (&wt);

  return salida;
}

/* ----------------------------------------------------------------------- */